@@ -1,20 +1,30 @@
-import time
-import json
 import asyncio
+import json
 import ssl
+import time
 from functools import partial
-from typing import Optional, Callable, NoReturn
+from typing import Callable, NoReturn, Optional
 
 import jwt
 import OpenSSL
 from h2.connection import H2Connection
-from h2.events import ResponseReceived, DataReceived, RemoteSettingsChanged,\
-    StreamEnded, ConnectionTerminated, WindowUpdated, SettingsAcknowledged
-from h2.exceptions import NoAvailableStreamIDError, FlowControlError
+from h2.events import (
+    ConnectionTerminated,
+    DataReceived,
+    RemoteSettingsChanged,
+    ResponseReceived,
+    SettingsAcknowledged,
+    StreamEnded,
+    WindowUpdated,
+)
+from h2.exceptions import FlowControlError, NoAvailableStreamIDError
 from h2.settings import SettingCodes
 
-from aioapns.common import NotificationResult, DynamicBoundedSemaphore,\
-    APNS_RESPONSE_CODE
+from aioapns.common import (
+    APNS_RESPONSE_CODE,
+    DynamicBoundedSemaphore,
+    NotificationResult,
+)
 from aioapns.exceptions import ConnectionClosed, ConnectionError
 from aioapns.logging import logger
 
@@ -24,7 +34,7 @@ class ChannelPool(DynamicBoundedSemaphore):
         super(ChannelPool, self).__init__(*args, **kwargs)
         self._stream_id = -1
 
-    async def acquire(self):
+    async def acquire(self) -> int:
         await super(ChannelPool, self).acquire()
         self._stream_id += 2
         if self._stream_id > H2Connection.HIGHEST_ALLOWED_STREAM_ID:
@@ -58,11 +68,11 @@ class JWTAuthorizationHeaderProvider(AuthorizationHeaderProvider):
         if not self.__header or self.__issued_at < now - self.TOKEN_TTL:
             self.__issued_at = int(now)
             token = jwt.encode(
-                payload={'iss': self.team_id, 'iat': self.__issued_at},
+                payload={"iss": self.team_id, "iat": self.__issued_at},
                 key=self.key,
-                algorithm='ES256',
-                headers={'kid': self.key_id},
-            ).decode('ascii')
+                algorithm="ES256",
+                headers={"kid": self.key_id},
+            )
             self.__header = f"bearer {token}"
         return self.__header
 
@@ -96,7 +106,7 @@ class H2Protocol(asyncio.Protocol):
             elif isinstance(event, SettingsAcknowledged):
                 pass
             else:
-                logger.warning('Unknown event: %s', event)
+                logger.warning("Unknown event: %s", event)
         self.flush()
 
     def flush(self):
@@ -110,13 +120,13 @@ class H2Protocol(asyncio.Protocol):
 
     def on_remote_settings_changed(self, changed_settings):
         for setting in changed_settings.values():
-            logger.debug('Remote setting changed: %s', setting)
+            logger.debug("Remote setting changed: %s", setting)
             if setting.setting == SettingCodes.MAX_CONCURRENT_STREAMS:
                 self.free_channels.bound = setting.new_value
 
     def on_stream_ended(self, stream_id):
         if stream_id % 2 == 0:
-            logger.warning('End stream: %d', stream_id)
+            logger.warning("End stream: %d", stream_id)
         self.free_channels.release()
 
     def on_connection_terminated(self, event):
@@ -124,15 +134,18 @@ class H2Protocol(asyncio.Protocol):
 
 
 class APNsBaseClientProtocol(H2Protocol):
-    APNS_SERVER = 'api.push.apple.com'
+    APNS_SERVER = "api.push.apple.com"
     INACTIVITY_TIME = 10
 
-    def __init__(self,
-                 apns_topic: str,
-                 loop: Optional[asyncio.AbstractEventLoop] = None,
-                 on_connection_lost: Optional[
-                     Callable[['APNsBaseClientProtocol'], NoReturn]] = None,
-                 auth_provider: Optional[AuthorizationHeaderProvider] = None):
+    def __init__(
+        self,
+        apns_topic: str,
+        loop: Optional[asyncio.AbstractEventLoop] = None,
+        on_connection_lost: Optional[
+            Callable[["APNsBaseClientProtocol"], NoReturn]
+        ] = None,
+        auth_provider: Optional[AuthorizationHeaderProvider] = None,
+    ):
         super(APNsBaseClientProtocol, self).__init__()
         self.apns_topic = apns_topic
         self.loop = loop or asyncio.get_event_loop()
@@ -152,29 +165,26 @@ class APNsBaseClientProtocol(H2Protocol):
         stream_id = await self.free_channels.acquire()
 
         headers = [
-            (':method', 'POST'),
-            (':scheme', 'https'),
-            (':path', '/3/device/%s' % request.device_token),
-            ('host', self.APNS_SERVER),
-            ('apns-id', request.notification_id),
-            ('apns-topic', self.apns_topic)
+            (":method", "POST"),
+            (":scheme", "https"),
+            (":path", "/3/device/%s" % request.device_token),
+            ("host", self.APNS_SERVER),
+            ("apns-id", request.notification_id),
+            ("apns-topic", self.apns_topic),
         ]
         if request.time_to_live is not None:
             expiration = int(time.time()) + request.time_to_live
-            headers.append(('apns-expiration', str(expiration)))
+            headers.append(("apns-expiration", str(expiration)))
         if request.priority is not None:
-            headers.append(('apns-priority', str(request.priority)))
+            headers.append(("apns-priority", str(request.priority)))
         if request.collapse_key is not None:
-            headers.append(('apns-collapse-id', request.collapse_key))
+            headers.append(("apns-collapse-id", request.collapse_key))
         if request.push_type is not None:
-            headers.append(('apns-push-type', request.push_type.value))
+            headers.append(("apns-push-type", request.push_type.value))
         if self.auth_provider:
-            headers.append(('authorization', self.auth_provider.get_header()))
+            headers.append(("authorization", self.auth_provider.get_header()))
 
-        self.conn.send_headers(
-            stream_id=stream_id,
-            headers=headers
-        )
+        self.conn.send_headers(stream_id=stream_id, headers=headers)
         try:
             data = json.dumps(request.message, ensure_ascii=False).encode()
             self.conn.send_data(
@@ -202,7 +212,8 @@ class APNsBaseClientProtocol(H2Protocol):
         if self.inactivity_timer:
             self.inactivity_timer.cancel()
         self.inactivity_timer = self.loop.call_later(
-            self.INACTIVITY_TIME, self.close)
+            self.INACTIVITY_TIME, self.close
+        )
 
     @property
     def is_busy(self):
@@ -212,7 +223,7 @@ class APNsBaseClientProtocol(H2Protocol):
         raise NotImplementedError
 
     def connection_lost(self, exc):
-        logger.debug('Connection %s lost!', self)
+        logger.debug("Connection %s lost!", self)
 
         if self.inactivity_timer:
             self.inactivity_timer.cancel()
@@ -226,8 +237,8 @@ class APNsBaseClientProtocol(H2Protocol):
         self.free_channels.destroy(closed_connection)
 
     def on_response_received(self, headers):
-        notification_id = headers.get(b'apns-id').decode('utf8')
-        status = headers.get(b':status').decode('utf8')
+        notification_id = headers.get(b"apns-id").decode("utf8")
+        status = headers.get(b":status").decode("utf8")
         if status == APNS_RESPONSE_CODE.SUCCESS:
             request = self.requests.pop(notification_id, None)
             if request:
@@ -235,14 +246,15 @@ class APNsBaseClientProtocol(H2Protocol):
                 request.set_result(result)
             else:
                 logger.warning(
-                    'Got response for unknown notification request %s',
-                    notification_id)
+                    "Got response for unknown notification request %s",
+                    notification_id,
+                )
         else:
             self.request_statuses[notification_id] = status
 
     def on_data_received(self, data, stream_id):
         data = json.loads(data.decode())
-        reason = data.get('reason', '')
+        reason = data.get("reason", "")
         if not reason:
             return
 
@@ -252,20 +264,26 @@ class APNsBaseClientProtocol(H2Protocol):
             if request:
                 # TODO: Теоретически здесь может быть ошибка, если нет ключа
                 status = self.request_statuses.pop(notification_id)
-                result = NotificationResult(notification_id, status,
-                                            description=reason)
+                result = NotificationResult(
+                    notification_id, status, description=reason
+                )
                 request.set_result(result)
             else:
-                logger.warning('Could not find request %s', notification_id)
+                logger.warning("Could not find request %s", notification_id)
         else:
-            logger.warning('Could not find notification by stream %s',
-                           stream_id)
+            logger.warning(
+                "Could not find notification by stream %s", stream_id
+            )
 
     def on_connection_terminated(self, event):
         logger.warning(
-            'Connection %s terminated: code=%s, additional_data=%s, '
-            'last_stream_id=%s', self, event.error_code,
-            event.additional_data, event.last_stream_id)
+            "Connection %s terminated: code=%s, additional_data=%s, "
+            "last_stream_id=%s",
+            self,
+            event.error_code,
+            event.additional_data,
+            event.last_stream_id,
+        )
         self.close()
 
 
@@ -275,25 +293,27 @@ class APNsTLSClientProtocol(APNsBaseClientProtocol):
     def close(self):
         if self.inactivity_timer:
             self.inactivity_timer.cancel()
-        logger.debug('Closing connection %s', self)
+        logger.debug("Closing connection %s", self)
         self.transport.close()
 
 
 class APNsProductionClientProtocol(APNsTLSClientProtocol):
-    APNS_SERVER = 'api.push.apple.com'
+    APNS_SERVER = "api.push.apple.com"
 
 
 class APNsDevelopmentClientProtocol(APNsTLSClientProtocol):
-    APNS_SERVER = 'api.development.push.apple.com'
+    APNS_SERVER = "api.development.push.apple.com"
 
 
 class APNsBaseConnectionPool:
-    def __init__(self,
-                 topic: Optional[str] = None,
-                 max_connections: int = 10,
-                 max_connection_attempts: Optional[int] = None,
-                 loop: Optional[asyncio.AbstractEventLoop] = None,
-                 use_sandbox: bool = False):
+    def __init__(
+        self,
+        topic: Optional[str] = None,
+        max_connections: int = 10,
+        max_connection_attempts: Optional[int] = None,
+        loop: Optional[asyncio.AbstractEventLoop] = None,
+        use_sandbox: bool = False,
+    ):
 
         self.apns_topic = topic
         self.max_connections = max_connections
@@ -315,10 +335,9 @@ class APNsBaseConnectionPool:
             connection.close()
 
     def discard_connection(self, connection):
-        logger.debug('Connection %s discarded', connection)
+        logger.debug("Connection %s discarded", connection)
         self.connections.remove(connection)
-        logger.info('Connection released (total: %d)',
-                    len(self.connections))
+        logger.info("Connection released (total: %d)", len(self.connections))
 
     async def acquire(self):
         for connection in self.connections:
@@ -334,17 +353,18 @@ class APNsBaseConnectionPool:
                 try:
                     connection = await self.create_connection()
                 except Exception as e:
-                    logger.error('Could not connect to server: %s', str(e))
+                    logger.error("Could not connect to server: %s", str(e))
                     self._lock.release()
                     raise ConnectionError()
                 self.connections.append(connection)
-                logger.info('Connection established (total: %d)',
-                            len(self.connections))
+                logger.info(
+                    "Connection established (total: %d)", len(self.connections)
+                )
                 self._lock.release()
                 return connection
             else:
                 self._lock.release()
-                logger.warning('Pool is busy, wait...')
+                logger.warning("Pool is busy, wait...")
                 while True:
                     await asyncio.sleep(0.01)
                     for connection in self.connections:
@@ -354,36 +374,50 @@ class APNsBaseConnectionPool:
     async def send_notification(self, request):
         failed_attempts = 0
         while True:
-            logger.debug('Notification %s: waiting for connection',
-                         request.notification_id)
+            logger.debug(
+                "Notification %s: waiting for connection",
+                request.notification_id,
+            )
             try:
                 connection = await self.acquire()
             except ConnectionError:
                 failed_attempts += 1
-                logger.warning('Could not send notification %s: '
-                               'ConnectionError', request.notification_id)
+                logger.warning(
+                    "Could not send notification %s: " "ConnectionError",
+                    request.notification_id,
+                )
 
-                if self.max_connection_attempts \
-                        and failed_attempts > self.max_connection_attempts:
-                    logger.error('Failed to connect after %d attempts.',
-                                 failed_attempts)
+                if (
+                    self.max_connection_attempts
+                    and failed_attempts > self.max_connection_attempts
+                ):
+                    logger.error(
+                        "Failed to connect after %d attempts.", failed_attempts
+                    )
                     raise
 
                 await asyncio.sleep(1)
                 continue
-            logger.debug('Notification %s: connection %s acquired',
-                         request.notification_id, connection)
+            logger.debug(
+                "Notification %s: connection %s acquired",
+                request.notification_id,
+                connection,
+            )
             try:
                 response = await connection.send_notification(request)
                 return response
             except NoAvailableStreamIDError:
                 connection.close()
             except ConnectionClosed:
-                logger.warning('Could not send notification %s: '
-                               'ConnectionClosed', request.notification_id)
+                logger.warning(
+                    "Could not send notification %s: " "ConnectionClosed",
+                    request.notification_id,
+                )
             except FlowControlError:
-                logger.debug('Got FlowControlError for notification %s',
-                             request.notification_id)
+                logger.debug(
+                    "Got FlowControlError for notification %s",
+                    request.notification_id,
+                )
                 await asyncio.sleep(1)
 
 
@@ -397,6 +431,7 @@ class APNsCertConnectionPool(APNsBaseConnectionPool):
         loop: Optional[asyncio.AbstractEventLoop] = None,
         use_sandbox: bool = False,
         no_cert_validation: bool = False,
+        ssl_context: Optional[ssl.SSLContext] = None,
     ):
 
         super(APNsCertConnectionPool, self).__init__(
@@ -408,14 +443,14 @@ class APNsCertConnectionPool(APNsBaseConnectionPool):
         )
 
         self.cert_file = cert_file
-        self.ssl_context = ssl.create_default_context()
+        self.ssl_context = ssl_context or ssl.create_default_context()
         if no_cert_validation:
             self.ssl_context.check_hostname = False
             self.ssl_context.verify_mode = ssl.CERT_NONE
         self.ssl_context.load_cert_chain(cert_file)
 
         if not self.apns_topic:
-            with open(self.cert_file, 'rb') as f:
+            with open(self.cert_file, "rb") as f:
                 body = f.read()
                 cert = OpenSSL.crypto.load_certificate(
                     OpenSSL.crypto.FILETYPE_PEM, body
@@ -428,25 +463,28 @@ class APNsCertConnectionPool(APNsBaseConnectionPool):
                 self.protocol_class,
                 self.apns_topic,
                 self.loop,
-                self.discard_connection
+                self.discard_connection,
             ),
             host=self.protocol_class.APNS_SERVER,
             port=self.protocol_class.APNS_PORT,
-            ssl=self.ssl_context
+            ssl=self.ssl_context,
         )
         return protocol
 
 
 class APNsKeyConnectionPool(APNsBaseConnectionPool):
-    def __init__(self,
-                 key_file: str,
-                 key_id: str,
-                 team_id: str,
-                 topic: str,
-                 max_connections: int = 10,
-                 max_connection_attempts: Optional[int] = None,
-                 loop: Optional[asyncio.AbstractEventLoop] = None,
-                 use_sandbox: bool = False):
+    def __init__(
+        self,
+        key_file: str,
+        key_id: str,
+        team_id: str,
+        topic: str,
+        max_connections: int = 10,
+        max_connection_attempts: Optional[int] = None,
+        loop: Optional[asyncio.AbstractEventLoop] = None,
+        use_sandbox: bool = False,
+        ssl_context: Optional[ssl.SSLContext] = None,
+    ):
 
         super(APNsKeyConnectionPool, self).__init__(
             topic=topic,
@@ -456,6 +494,8 @@ class APNsKeyConnectionPool(APNsBaseConnectionPool):
             use_sandbox=use_sandbox,
         )
 
+        self.ssl_context = ssl_context or ssl.create_default_context()
+
         self.key_id = key_id
         self.team_id = team_id
 
@@ -464,9 +504,7 @@ class APNsKeyConnectionPool(APNsBaseConnectionPool):
 
     async def create_connection(self):
         auth_provider = JWTAuthorizationHeaderProvider(
-            key=self.key,
-            key_id=self.key_id,
-            team_id=self.team_id
+            key=self.key, key_id=self.key_id, team_id=self.team_id
         )
         _, protocol = await self.loop.create_connection(
             protocol_factory=partial(
@@ -478,6 +516,6 @@ class APNsKeyConnectionPool(APNsBaseConnectionPool):
             ),
             host=self.protocol_class.APNS_SERVER,
             port=self.protocol_class.APNS_PORT,
-            ssl=True,
+            ssl=self.ssl_context,
         )
         return protocol
